package livework;
/* (c) Alex McLean 2006
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

import java.lang.System;

/**
 * Threaded object to handle reasonably accurate timing for an  application.<br />   When instantiated a Clock goes at a default rate of 120  'ticks' per minute, changeable via the setTicksPerMinute()  method.  Every tick each member object in the triggerables  array has its trigger() method called.
 * @author  Alex Mclean - ma503am@gold.ac.uk
 */
public class Clock implements Runnable {
	/** Maximum number of objects this object can handle. */
	public final static int maxTriggerables = 256;
	/** milliseconds per 'tick' */
	private double interval; // milliseconds
	/** the time the clock starts */
	private long startTime;
	/** number of ticks since the clock started */
	private int ticks;
	/** number of objects in this object's triggerable array */
	private int triggerableCount = 0;
	/**
	 * objects to trigger() every 'tick'
	 * @uml.property  name="triggerables"
	 * @uml.associationEnd  multiplicity="(0 -1)"
	 */
	private Triggerable[] triggerables = new Triggerable[maxTriggerables];
	
	/**
	 * Class constructor
	 */
	public Clock() {
		setTicksPerMinute(300);
	}
	
	/**
	 * @param tpm number of ticks per minute
	 */
	void setTicksPerMinute(double tpm) {
		// convert to interval in milliseconds
		interval = 1000.0 / (tpm / 60.0);
	}
	
	double getTicksPerMinute() {
		return((1000.0 / interval) * 60.0);
	}
	
	/**
	 * Adds an object to receive clock ticks via it's triggerable() method.  
	 * The object must implement the Triggerable interface.
	 * @param triggerable
	 * @return true if the object was successfully added
	 */
	public boolean addTriggerable(Triggerable triggerable) {
		boolean result = false;
		if (triggerableCount < maxTriggerables) {
			triggerables[triggerableCount++] = triggerable;
		}
		return(result);
	}
	
	/**
	 * Thread starting point.  Just calls the tickLoop() method.
	 */
	public void run() {
		tickLoop();
	}

	/**
	 * called once per tick by the tickLoop().  Calls
	 * the trigger() array on each object in the triggerable 
	 * array
	 */
	private void tick() {
		++ticks;
		if (triggerableCount > 0) {
			for (int i = 0; i < triggerableCount; ++i) {
				triggerables[i].trigger();
			}
		}
	}

	/**
	 * The thread loop.  Calls tick() at the defined rate.  Uses
	 * regulates itself using java.lang.nanoTime() for accuracy,
	 * particularly on linux based systems.
	 */
	private void tickLoop() {
		startTime = System.nanoTime();
		ticks = 0;
		while(true) {
			// tick!
			tick();
			
			// calculate target and elapsed afresh each time to avoid 
			// accumulating errors
			double target = ticks * interval;
			double elapsed = (System.nanoTime() - startTime) / 1000000.0;
			
			if (target > elapsed) {
				try {
					Thread.sleep((long) (target - elapsed));
				}
				catch (InterruptedException e) {
					// ignore - don't worry about super precise timing for now,
					// the error will be absorbed next loop anyway
				}
			}
		}
	}
}

