package livework;
/* (c) Alex McLean 2006
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

import java.lang.reflect.*;
import java.io.*;
import java.util.*;

import livework.Settings;

import com.sun.tools.javac.*;

/**
 * <p>This class is the heart of this project.  Refer to the project report for
 * information about how I arrived at this implementation.</p>
 * <p>It employs a number of techniques towards the end of allowing a
 * programmer to make changes to the behaviour of objects by changing
 * their classes.  This isn't actually possible in Java yet, you cannot add
 * a method to a class that has already been loaded.  This class achieves an
 * approximate effect by using the following smoke and mirrors:
 * </p>
 * <ul>
 *   <li>A proxy object - the LiveProxy class acts as a proxy for the object that is being 
 *       live-edited.  This allows all calls to the object to be intercepted.  When a call 
 *       is made to the object via the proxy, before passing the call on, the proxy checks
 *       the proxied object's class sourcecode for changes.  If the sourcecode has been 
 *       changed, the class is dynamically compiled and loaded into a fresh package.
 *   </li>
 *   <li>Compilation - this is accomplished by calling javac via the 
 *       "com.sun.tools.javac.Main" package.  Before compilation the source is modified 
 *       to include a unique package statement.  This is necessary because a class can't be
 *       loaded twice into the same package.
 *   </li>
 *   <li>Dynamic loading - Because this class extends the ClassLoader class, it can load
 *       classes at runtime.  Once the class of the proxied object has been successfully
 *       compiled it loads the class ready to be instantiated.
 *   </li>
 *   <li>Reflection - Because we can't really change the class of the proxied object (simplification -
 *       see project report for the full story), we've loaded the class into a fresh package.  This
 *       way it can keep its original unqualified classname and therefore the names of its 
 *       constructors and so on.  Now an object can't be moved from one class to another, so we
 *       do the next best thing - make a new object, copy across all the fields we can, then forget 
 *       about the original.  This is akin to serialising an object, then deserialising it into a 
 *       different class, all in one go.  Not all the fields will be copyable, some fields might
 *       have been deleted by the programmer, or had been changed to a type incompatible with the
 *       original.  All this is made possible, perhaps even <em>easy</em>, by Java reflection.
 *       The classes, methods are all accessible as objects.
 *   </li>
 *   <li>Exception handling - if any compilation errors are caught, execution continues with the
 *       original version of the object intact.
 *   </li>
 * </ul>
 * <p>
 * So that's a short overview, for further details refer to the methods.
 * </p>
 * <p>
 * Note that a LiveProxy object also starts an Editor to edit the sourcecode
 * of the class that's being proxied.  Once the programmer saves their changes
 * to the source, the next call to the object via the proxy will cause the object
 * to be recompiled as above.
 * </p>
 * @author Alex McLean - ma503am@gold.ac.uk
 */

public class LiveProxy extends ClassLoader implements InvocationHandler {
	/** Count for the number of times this class has been recompiled, used
	 * for constructing unique package names. */
	private int count;
	/** The filename of the sourcecode for the class.  It lives within the 
	 * folder defined by the 'liveclasses' setting.
	 * @see Settings
	 */
	private String filename;
	/** Records the last modified time of the sourcefile, so we can tell
	 * when it is modified again
	 */
	private long lastModified;
	/** Represents the fully qualified filename of the sourcefile */
	private File sourcecode;	
	/** The compiled target object */
	private Object target = null;

	/**
	 * Instantiates itself with the given sourcefile, instantiating the
	 * target object in the process, and returns a proxy class
	 * that acts as the target object but all accesses are passed to this
	 * object's invoke() method
	 * @param filename sourcefile of the class for the target object
	 * @return the proxy object
	 */
	public static Object createProxy(String filename) {
		LiveProxy liveProxy = new LiveProxy(filename);
		Object object = liveProxy.target;
		
		return Proxy.newProxyInstance(object.getClass().getClassLoader(),
									  object.getClass().getInterfaces(),
									  liveProxy
									 );
	}
	
	/**
	 * The following method was taken from public domain book example - 
	 * Forman and Forman, Java Reflection in Action.<br />
	 * Returns an array of fields for the given class
	 */
	static Field[] getInstanceVariables(Class cls) {
    	List accum = new LinkedList();
    	while (cls != null) {
    		Field[] fields = cls.getDeclaredFields();
    	    for (int i=0; i<fields.length; i++) {
    	    	if (!Modifier.isStatic(fields[i].getModifiers())) {
    	    		accum.add(fields[i]);
    	    	}	
    	    }
    	    cls = cls.getSuperclass();
    	}
    	Field[] retvalue = new Field[accum.size()];
    	return (Field[]) accum.toArray(retvalue);
    }
	
	/**
	 * The constructor for this class.  It compiles and instantiates
	 * the class identified by the given filename.  If the file doesn't
	 * exist, it is created with some default sourcecode.
	 * @param filename
	 */
	private LiveProxy(String filename) {
		this.filename = filename;
		sourcecode = new File(classdir() + "/" + filename + ".java");		
		
		if (! sourcecode.exists()) {
			defaultSourcecode();
		}
		
		lastModified = sourcecode.lastModified();
		try {
			target = loadObject();
			initGlobals(target);	
		}
		catch (CompileException e) {
			// couldn't compile the code, don't have any previous object to
			// fall back on, so use a minimal object that does nothing as the 
			// target for now.
			target = new MinimalTrigger();
		}
		catch (Throwable e) {
			e.printStackTrace();
		};
		
		new Editor(sourcecode);
	}
	
	/**
	 * Increases the count field by one and saves it to disk.  This
	 * means that each edit of the java source is saved to a fresh packages 
	 * across multiple invocations of this class.  This isn't necessary 
	 * at the moment but future versions of this application might allow
	 * the programmer to jump easily jump back to editing and reloading 
	 * previous versions of the class.
	 */
	private synchronized void bumpCount () {
		String scount = "";
		String path = classdir() + "/" + filename + ".count";
		int foo;
		try {
			FileReader fileReader = new FileReader(path);
			while ((foo = fileReader.read()) != -1) {
				scount = scount + (char) foo;
			}
			fileReader.close();
			
			count = Integer.parseInt(scount);
			count++;
		}
		catch (Exception e) {
			count = 0;
		}
		
		try {
			FileWriter fileWriter = new FileWriter(path);
			fileWriter.write(Integer.toString(count));
			fileWriter.close();
		}
		catch (IOException e) {
			// hmm...
		}
	}
	
	/**
	 * Retrieves from the user settings the folder where the sourcefiles 
	 * to be edited are located.
	 */
	private String classdir() {
		return(Settings.sget("liveClasspath"));
	}

	/**
	 * Copies the sourcefile to a fresh package and compiles it there
	 * @return the name of the new class qualified by the new package
	 * @throws CompileException thrown when the compile wasn't successful.
	 *         Compile errors are sent to standard error, as captured by
	 *         the editor and presented to the programmer to meditate
	 *         over.
	 */
	private String copyClass() throws CompileException {
		bumpCount();
		String newPath = classdir() + "/" + "v" + count;
		java.io.File dir = new File(newPath);
		if (! dir.exists()) {
			dir.mkdir();
		}
		
		String origFile = sourcecode.getAbsolutePath();
		String newFile = newPath + "/" + filename + ".java";
		
		char[] buf = new char[1024];
		try {
			FileReader fileReader = new FileReader(origFile);
			FileWriter fileWriter = new FileWriter(newFile);

			String packageStatement = "package v" + count + ";\n";
			fileWriter.write(packageStatement.toCharArray());

			int read;
			while ((read = fileReader.read(buf, 0, 1024)) != -1) {
				fileWriter.write(buf, 0, read);
			}
			fileReader.close();
			fileWriter.close();
		}
		catch (Exception e) {
			count = 0;
		}

		String args[] = {"-d", classdir(), newFile};
		if (Main.compile(args) != 0) {
			throw(new CompileException("couldn't compile '" + newFile + "'"));
		}
		else {
			System.err.println("Compiled OK");
		}

		return("v" + count + "." + filename);
	}
	
	
	private void copyFields(Object newTarget, Object source) {
		
		Class sourceclass = source.getClass();

		Field[] fields = getInstanceVariables(sourceclass);
		for (int i = 0; i < fields.length; ++i) {
			if (!Modifier.isPublic(fields[i].getModifiers())) {
				fields[i].setAccessible(true);
			}
			String fieldName = fields[i].getName();
			try {
				Field newField = findField(newTarget.getClass(), fieldName);
				if (!Modifier.isPublic(newField.getModifiers())) {
					newField.setAccessible(true);
				}
				Object tmp = fields[i].get(source);
				newField.set(newTarget, tmp);
			}
			catch (IllegalArgumentException e) {
				System.out.println("couln't set " + fieldName + "; " + e);
			}			
			catch (Exception e) {
				System.out.println("couln't set " + fieldName + "; " + e);
			}
		}
	}
	
	private void defaultSourcecode() {
		sourcecode.getParentFile().mkdirs();
		try {
			FileOutputStream output = new FileOutputStream(sourcecode);
			PrintWriter writer = new PrintWriter(new OutputStreamWriter(output));

			String code = ("import livework.*;"
						 + "import java.awt.*;\n"
						 + "import java.lang.Math;\n"
						 + "public class Bare implements Triggerable {\n"
						 + "    Nourathar nourathar = null;\n"
						 + "    int foo = 0;\n"
						 + "    public void trigger() {\n"
						 + "        if (foo % 16 == 0) {\n"
						 + "            // color, envelope name, duration, x, y, size\n"
						 + "            nourathar.play(Color.BLUE, \"triangular\", 30, 0.2, 0.1, 0.5);\n"
						 + "        }\n"
						 + "        if (foo % 3 == 0) {\n"
						 + "            // color, envelope name, duration, x, y, size\n"
						 + "            nourathar.play(Color.ORANGE, \"percussive\", 8, 0.3, 0.25, 0.1);\n"
						 + "        }\n"
						 + "	if (foo % 4 == 0) {\n"
						 + "            // color, envelope name, duration, x, y, size\n"
						 + "            nourathar.play(Color.RED, \"percussive\", 8, 0.5, 0.25, 0.1);\n"
						 + "        }\n"
						 + "        if (foo % 5 == 0) {\n"
						 + "            nourathar.play(Color.GREEN, \"percussive\", 10, 0.4, 0.45, 0.1);\n"
						 + "        }\n"
						 + "        //int x = (int) Math.round(Math.sin(foo / 50.0) * 150) + 200;\n"
						 + "        //int y = (int) Math.round(Math.cos(foo / 50.0) * 150) + 200;\n"
						 + "        //nourathar.setLocation(x, y);\n"
						 + "        foo++;\n"
						 + "    }\n"
						 + "    public void initGlobals() {\n"
						 + "        if (nourathar == null) {\n"
						 + "            nourathar = new Nourathar();\n"
						 + "        }\n"
						 + "    }\n"
						 + "}\n"
		                   );
			writer.print(code);
			writer.close();
		}
		catch(FileNotFoundException e) {
			Util.error("System error", "Couldn't write to file '" + sourcecode.getAbsoluteFile() + "'", e);
		};
	}

	Field findField(Class myClass, String fieldName) {
		Field[] fields = getInstanceVariables(myClass);
		Field result = null;
		for (int i = 0; i < fields.length; ++i) {
			if (fields[i].getName() == fieldName) {
				result = fields[i];
				break;
			}
		}
		return(result);
	}
	
	/**
	 * The following method was taken from a public domain example from the
	 * book - 
	 * Forman and Forman, Java Reflection in Action.<br />
	 * It returns the byte compiled class data for the given file within the 
	 * given directory.
	 */
	private byte[] getClassData(String directory, String name) {
		String classFile = directory + "/" + name.replace('.', '/') + ".class";
		int classSize = (new Long((new File(classFile)).length())).intValue();
		byte[] buf = new byte[classSize];
		try {
			FileInputStream filein = new FileInputStream(classFile);
			classSize = filein.read(buf);
			filein.close();
		}
		catch (FileNotFoundException e) {
			return null;
		}
		catch (IOException e) {
			return null;
		}
		return buf;
	}
	
	/**
	 * This method sidesteps a big problem with this whole approach - field initialisation.
	 * When you instantiate an object, fields are often initialised, ie:<br />
	 * <code> int foo = 40;</code>
	 * <br />
	 * Now if we changed that value to 41 and triggered a reload of the class, the field
	 * would get initialised to 41, but then overwritten back to 40 by the copyFields()
	 * method.  This might be what the programmer wanted to happen, but then again, might 
	 * not.  Further, the initialisation might do something that the programmer really
	 * only wants to happen once, not every time the program is compiled.
	 * <br />To allow the programmer full control over field initialisation, this method 
	 * is called <em>after</em> copyFields().  It reflectively looks for a method on
	 * the new target object called "initGlobals()".  If it finds one, it invokes (calls) it.  If
	 * it doesn't, it does nothing.<br />
	 * The programmer can therefore write code in such a method that initialises fields
	 * exactly as s/he wants.
	 * @param object The new target object
	 * @throws Throwable An exception raised by a initGlobals method on the new target object.
	 */
	private void initGlobals(Object object) throws Throwable {
		try {
			Class[] args = {};
			Method method = object.getClass().getMethod("initGlobals", args);
			method.invoke(object, null);
		}
		catch (NoSuchMethodException e) {
			// ignore exception, the initGlobals() method is optional
		}
		catch (IllegalAccessException e) {
			// ignore exception, if they want to hide the init method,
			// don't run it.  print out a warning though
			System.err.println("Couldn't access init() method");
		}
		catch (InvocationTargetException e) {
			throw(e.getCause());
		}
	}
	
	/**
	 * Intercepts calls to the target object via this proxy object.  If necessary, it reloads 
	 * the target object invoking the method upon it.
	 */
	public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
		if (lastModified < sourcecode.lastModified()) {
			reloadTarget();
			lastModified = sourcecode.lastModified();
		}
		Object result = null;
		try {
			result = method.invoke(target, args);
		}
		catch (InvocationTargetException e) {
			throw e.getCause();
		}
		return(result);
	}
	
	/**
	 * Loads the class and instantiates a new target object from it.
	 * @return the new target object
	 * @throws CompileException thrown when the compile failed
	 */
	private Object loadObject() throws CompileException {
		String newClassfile = copyClass();
		
		byte[] buf = getClassData(classdir(), newClassfile);
		Object newTarget = null;
		try {
			if (buf != null) {
				Class newClass = defineClass(newClassfile, buf, 0, buf.length);
				if (newClass != null) {
					newTarget = newClass.newInstance();
				}
				else {
					System.out.println("defineClass didn't work");
				}
			}
			else {
				System.out.println("buf was null");
			}
		}
		catch (IllegalAccessException e) {
			System.out.println("illegal access");
			newTarget = null;
		}
		catch (InstantiationException e) {
			System.out.println("instantiation");
			newTarget = null;
		}
		return(newTarget);
	}
	
	/**
	 * Reloads the target object from its sourcecode.
	 */
	public synchronized void reloadTarget() {
		try {
			Object newTarget = loadObject();
			copyFields(newTarget, target);
			initGlobals(newTarget);	
			target = newTarget;
		}
		catch (CompileException e) {
			System.err.println(e.getMessage());
		}
		catch (Throwable e) {
			e.printStackTrace();
		}
	}
}

/**
 * A simple exception to represent a compile error.
 * @author Alex McLean - ma503am@gold.ac.uk
 */
class CompileException extends Exception {
	// I'm not going to serialize this class, but this avoids a compiler warning
	private static final long serialVersionUID = 1;
	CompileException() {
		super();
	}
	CompileException(String message) {
		super(message);
	}
}

/**
 * A class to instantiate a simple object from as a placeholder if a target object
 * couldn't be loaded at runtime.  This is only used if an object doesn't compile
 * when the LiveProxy is first instantiated, where a previous compile isn't available
 * to fall back on.
 * @author Alex McLean - ma503am@gold.ac.uk
 */
class MinimalTrigger implements Triggerable {
	public void trigger() {
		
	}
}
