package livework;
/* (c) Alex McLean 2006
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

import java.awt.*;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Iterator;
import javax.swing.*;

/**
 * A frame for producing simple live video animations.  Intended as a demonstration for the
 * rest of this application but explores some ideas inspired by Mary Hallock
 * Greenwelt's concept of "Nourathar" - the playing of colour transitions as
 * a performance.<br />
 * I implemented an offscreen buffer to avoid flicker, but since then have readd that java2d
 * can do this internally.  If I develop this further I'll be sure to simplify things by 
 * removing the manual offscreen buffer.
 * @author Alex McLean - ma503am@gold.ac.uk
 */

public class Nourathar extends JFrame implements Runnable {
	/** I'm not going to serialize this class, but this avoids a compiler warning */
	private static final long serialVersionUID = 1;

	/** to contain the currently active NourtharColor objects */
	private LinkedList colors = new LinkedList();
	/** a lookup table for available envelopes to apply to the brightness levels of the NouratharColor objects */
	private Hashtable envelopes = null;
	/** The graphics context of the offscreen buffer */
	private Graphics2D g2d = null;
	/** The offscreen buffer */
	private Image offscreenBuffer = null;
	
	/**
	 * Contructor for this class.  Initialises frame, offscreen buffer, envelopes and starts 
	 * the thread for the object.
	 */
	public Nourathar () {
		super("Nourathar");
        setSize(new Dimension(600, 480));
        setVisible(true);
        offscreenBuffer = getContentPane().createImage(getContentPane().getWidth(),
        		                                       getContentPane().getHeight()
        		                                      );
        
        g2d = (Graphics2D) offscreenBuffer.getGraphics();
        g2d.setBackground(Color.BLACK);
        initEnvelopes();
		new Thread(this).start();
	}

	/**
	 * Initialises the envelopes.  
	 */
	void initEnvelopes() {
    	envelopes = new Hashtable();
    	
    	// percusive
    	float[] percussive= {0, 1, 0.7f, 0.5f, 0.3f, 0.3f, 0.2f, 0.1f, 0};
    	envelopes.put("percussive", new Envelope(percussive));
    	
    	// triangular
    	float[] triangular = {0, 1, 0};
    	envelopes.put("triangular", new Envelope(triangular));
    }
	
	/**
	 * Paints the offscreen buffer.
	 */
    public void paint(Graphics g) {
        //super.paint(g);
        Graphics2D g2 = (Graphics2D) getContentPane().getGraphics();
        g2.drawImage(offscreenBuffer, 0, 0, null);
    }    

    /**
     * Method for 'playing' the nourathar.  Causes a NouratharColor visual element to appear 
     * of the given 
     * colour, location and size for the given length of time, with the brightness modulated
     * by the given envelope.  The parameters are passed straight to the constructor of the
     * NouratharColor class, apart from the 'x', 'y' and 'size' parameters, which are given 
     * as proportion of the content pane size and converted to pixel values. 
     * @param color Colour of the element at it's maximum brightness
     * @return  The created NouratharColor object
     */
    public synchronized NouratharColor play(Color color, String envelopeName, int ttl, double x, double y, double size) {
    	Envelope envelope = (Envelope) envelopes.get(envelopeName);

    	if (envelope == null) {
    		envelope = (Envelope) envelopes.get("percussive");
    	}
    	
    	x *= getContentPane().getWidth();
    	y *= getContentPane().getHeight();
    	size *= getContentPane().getWidth();
 
  		NouratharColor result = new NouratharColor(color, envelope, ttl, x, y, size);
    	colors.add(result);
    	
    	return(result);
    }
    
    /**
     * Draws the active nouratharColor visual elements on the offscreen buffer
     */
    synchronized void render() {
		g2d.setColor(Color.BLUE);
		g2d.clearRect(0, 0, getContentPane().getWidth(), getContentPane().getHeight());
		Iterator i = colors.iterator();
		LinkedList toRemove = new LinkedList();
		
		while (i.hasNext()) {
			NouratharColor nouratharColor = (NouratharColor) i.next();

			nouratharColor.draw(g2d);
			if (nouratharColor.isExpired()) {
				toRemove.add(nouratharColor);
			}
		}
		i = toRemove.iterator();
		while (i.hasNext()) {
			NouratharColor nouratharColor = (NouratharColor) i.next();
			colors.remove(nouratharColor);
		}
	}
    
    /** 
     * Thread loop for this class.  Renders and paints the window once per
     * 50 milliseconds
     */
    public void run() {
        while(true) {
        	try {
        		render();
            	repaint();
        		Thread.sleep(50);
        	}
        	catch (InterruptedException e) {
        	    // don't worry about timing for now
        	}
        }
    }
}
